@@ -1,18 +1,18 @@
 /*
- * Unix SMB/CIFS implementation. 
+ * Unix SMB/CIFS implementation.
  * Copyright (C) Jeremy Allison 1995-1998
  * Copyright (C) Tim Potter     2001
- * 
+ *
  * This program is free software; you can redistribute it and/or modify it
  * under the terms of the GNU General Public License as published by the
  * Free Software Foundation; either version 3 of the License, or (at your
  * option) any later version.
- * 
+ *
  * This program is distributed in the hope that it will be useful, but WITHOUT
  * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
  * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
  * more details.
- * 
+ *
  * You should have received a copy of the GNU General Public License along with
  * this program; if not, see <http://www.gnu.org/licenses/>.  */
 
@@ -26,6 +26,8 @@
 #include "cmdline_contexts.h"
 #include "passwd_proto.h"
 #include "lib/util/string_wrappers.h"
+#include "lib/param/param.h"
+#include "lib/util/memcache.h"
 
 /*
  * Next two lines needed for SunOS and don't
@@ -87,7 +89,8 @@ static void set_line_buffering(FILE *f)
  Process command line options
  ******************************************************************/
 
-static int process_options(int argc, char **argv, int local_flags)
+static int process_options(int argc, char **argv, int local_flags,
+			   struct loadparm_context *lp_ctx)
 {
 	int ch;
 	const char *configfile = get_dyn_CONFIGFILE();
@@ -156,10 +159,10 @@ static int process_options(int argc, char **argv, int local_flags)
 			fstrcpy(ldap_secret, optarg);
 			break;
 		case 'R':
-			lp_set_cmdline("name resolve order", optarg);
+			lpcfg_set_cmdline(lp_ctx, "name resolve order", optarg);
 			break;
 		case 'D':
-			lp_set_cmdline("log level", optarg);
+			lpcfg_set_cmdline(lp_ctx, "log level", optarg);
 			break;
 		case 'U': {
 			got_username = True;
@@ -200,7 +203,7 @@ static int process_options(int argc, char **argv, int local_flags)
 	}
 
 	if (!lp_load_global(configfile)) {
-		fprintf(stderr, "Can't load %s - run testparm to debug it\n", 
+		fprintf(stderr, "Can't load %s - run testparm to debug it\n",
 			configfile);
 		exit(1);
 	}
@@ -290,8 +293,8 @@ static NTSTATUS password_change(const char *remote_mach,
  Store the LDAP admin password in secrets.tdb
  ******************************************************************/
 static bool store_ldap_admin_pw (char* pw)
-{	
-	if (!pw) 
+{
+	if (!pw)
 		return False;
 
 	if (!secrets_init())
@@ -346,9 +349,9 @@ static int process_root(int local_flags)
 	 * Ensure both add/delete user are not set
 	 * Ensure add/delete user and either remote machine or join domain are
 	 * not both set.
-	 */	
-	if(((local_flags & (LOCAL_ADD_USER|LOCAL_DELETE_USER)) == (LOCAL_ADD_USER|LOCAL_DELETE_USER)) || 
-	   ((local_flags & (LOCAL_ADD_USER|LOCAL_DELETE_USER)) && 
+	 */
+	if(((local_flags & (LOCAL_ADD_USER|LOCAL_DELETE_USER)) == (LOCAL_ADD_USER|LOCAL_DELETE_USER)) ||
+	   ((local_flags & (LOCAL_ADD_USER|LOCAL_DELETE_USER)) &&
 		(remote_machine != NULL))) {
 		usage();
 	}
@@ -362,7 +365,7 @@ static int process_root(int local_flags)
 	if (!user_name[0] && (pwd = getpwuid_alloc(talloc_tos(), geteuid()))) {
 		fstrcpy(user_name, pwd->pw_name);
 		TALLOC_FREE(pwd);
-	} 
+	}
 
 	if (!user_name[0]) {
 		fprintf(stderr,"You must specify a username\n");
@@ -478,7 +481,7 @@ static int process_root(int local_flags)
 					     local_flags))) {
 		result = 1;
 		goto done;
-	} 
+	}
 
 	if(remote_machine) {
 		printf("Password changed for user %s on %s.\n", user_name, remote_machine );
@@ -558,7 +561,7 @@ static int process_nonroot(int local_flags)
 	 * A non-root user is always setting a password
 	 * via a remote machine (even if that machine is
 	 * localhost).
-	 */	
+	 */
 
 	load_interfaces(); /* Delayed from main() */
 
@@ -615,11 +618,20 @@ static int process_nonroot(int local_flags)
  Start here.
 **********************************************************/
 int main(int argc, char **argv)
-{	
+{
 	TALLOC_CTX *frame = talloc_stackframe();
+	struct loadparm_context *lp_ctx = NULL;
+	struct memcache *mcache = NULL;
 	int local_flags = 0;
 	int ret;
 
+	mcache = memcache_init(NULL, 0);
+	if (mcache == NULL) {
+		fprintf(stderr, "%s: memcache_init failed\n", __location__);
+		return 1;
+	}
+	memcache_set_global(mcache);
+
 #if defined(HAVE_SET_AUTH_PARAMETERS)
 	set_auth_parameters(argc, argv);
 #endif /* HAVE_SET_AUTH_PARAMETERS */
@@ -630,18 +642,17 @@ int main(int argc, char **argv)
 
 	smb_init_locale();
 
-	local_flags = process_options(argc, argv, local_flags);
+	lp_ctx = loadparm_init_s3(frame, loadparm_s3_helpers());
+	if (lp_ctx == NULL) {
+		fprintf(stderr,
+			"Failed to initialise the global parameter structure.\n");
+		return 1;
+	}
+
+	local_flags = process_options(argc, argv, local_flags, lp_ctx);
 
 	setup_logging("smbpasswd", DEBUG_STDERR);
 
-	/*
-	 * Set the machine NETBIOS name if not already
-	 * set from the config file. 
-	 */ 
-
-	if (!init_names())
-		return 1;
-
 	/* Check the effective uid - make sure we are not setuid */
 	if (is_setuid_root()) {
 		fprintf(stderr, "smbpasswd must *NOT* be setuid root.\n");
@@ -659,6 +670,9 @@ int main(int argc, char **argv)
 	} else {
 		ret = process_nonroot(local_flags);
 	}
+
+	gfree_all();
+
 	TALLOC_FREE(frame);
 	return ret;
 }
