@@ -1,26 +1,33 @@
-/* 
+/*
    Unix SMB/CIFS implementation.
    Samba utility functions
 
    Copyright (C) Jelmer Vernooij <jelmer@samba.org> 2008-2010
-   
+
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 3 of the License, or
    (at your option) any later version.
-   
+
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
-   
+
    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
-#include <Python.h>
+#include "lib/replace/system/python.h"
+#include "gen_ndr/conditional_ace.h"
 #include "py3compat.h"
+#include "libcli/security/sddl.h"
 #include "libcli/security/security.h"
 
+
+/* Set up in py_mod_security_patch() */
+static PyObject *PyExc_SDDLValueError = NULL;
+
+
 static void PyType_AddMethods(PyTypeObject *type, PyMethodDef *methods)
 {
 	PyObject *dict;
@@ -30,11 +37,11 @@ static void PyType_AddMethods(PyTypeObject *type, PyMethodDef *methods)
 	dict = type->tp_dict;
 	for (i = 0; methods[i].ml_name; i++) {
 		PyObject *descr;
-		if (methods[i].ml_flags & METH_CLASS) 
+		if (methods[i].ml_flags & METH_CLASS)
 			descr = PyCFunction_New(&methods[i], (PyObject *)type);
-		else 
+		else
 			descr = PyDescr_NewMethod(type, &methods[i]);
-		PyDict_SetItemString(dict, methods[i].ml_name, 
+		PyDict_SetItemString(dict, methods[i].ml_name,
 				     descr);
 		Py_CLEAR(descr);
 	}
@@ -67,7 +74,6 @@ static PyObject *py_dom_sid_split(PyObject *py_self, PyObject *args)
 	return Py_BuildValue("(OI)", py_domain_sid, rid);
 }
 
-#if PY_MAJOR_VERSION >= 3
 static PyObject *py_dom_sid_richcmp(PyObject *py_self, PyObject *py_other, int op)
 {
 	struct dom_sid *self = pytalloc_get_ptr(py_self), *other;
@@ -92,25 +98,6 @@ static PyObject *py_dom_sid_richcmp(PyObject *py_self, PyObject *py_other, int o
 	Py_INCREF(Py_NotImplemented);
 	return Py_NotImplemented;
 }
-#else
-static int py_dom_sid_cmp(PyObject *py_self, PyObject *py_other)
-{
-	struct dom_sid *self = pytalloc_get_ptr(py_self), *other;
-	int val;
-
-	other = pytalloc_get_ptr(py_other);
-	if (other == NULL)
-		return -1;
-
-	val =  dom_sid_compare(self, other);
-	if (val > 0) {
-		return 1;
-	} else if (val < 0) {
-		return -1;
-	}
-	return 0;
-}
-#endif
 
 static PyObject *py_dom_sid_str(PyObject *py_self)
 {
@@ -139,7 +126,8 @@ static int py_dom_sid_init(PyObject *self, PyObject *args, PyObject *kwargs)
 		return -1;
 
 	if (str != NULL && !dom_sid_parse(str, sid)) {
-		PyErr_SetString(PyExc_TypeError, "Unable to parse string");
+		PyErr_Format(PyExc_ValueError,
+			     "Unable to parse string: '%s'", str);
 		return -1;
 	}
 
@@ -159,11 +147,7 @@ static void py_dom_sid_patch(PyTypeObject *type)
 	type->tp_init = py_dom_sid_init;
 	type->tp_str = py_dom_sid_str;
 	type->tp_repr = py_dom_sid_repr;
-#if PY_MAJOR_VERSION >= 3
 	type->tp_richcompare = py_dom_sid_richcmp;
-#else
-	type->tp_compare = py_dom_sid_cmp;
-#endif
 	PyType_AddMethods(type, py_dom_sid_extra_methods);
 }
 
@@ -175,12 +159,13 @@ static PyObject *py_descriptor_sacl_add(PyObject *self, PyObject *args)
 	NTSTATUS status;
 	struct security_ace *ace;
 	PyObject *py_ace;
+	Py_ssize_t idx = -1;
 
-	if (!PyArg_ParseTuple(args, "O", &py_ace))
+	if (!PyArg_ParseTuple(args, "O|n", &py_ace, &idx))
 		return NULL;
 
 	ace = pytalloc_get_ptr(py_ace);
-	status = security_descriptor_sacl_add(desc, ace);
+	status = security_descriptor_sacl_insert(desc, ace, idx);
 	PyErr_NTSTATUS_IS_ERR_RAISE(status);
 	Py_RETURN_NONE;
 }
@@ -191,13 +176,14 @@ static PyObject *py_descriptor_dacl_add(PyObject *self, PyObject *args)
 	NTSTATUS status;
 	struct security_ace *ace;
 	PyObject *py_ace;
+	Py_ssize_t idx = -1;
 
-	if (!PyArg_ParseTuple(args, "O", &py_ace))
+	if (!PyArg_ParseTuple(args, "O|n", &py_ace, &idx))
 		return NULL;
 
 	ace = pytalloc_get_ptr(py_ace);
 
-	status = security_descriptor_dacl_add(desc, ace);
+	status = security_descriptor_dacl_insert(desc, ace, idx);
 	PyErr_NTSTATUS_IS_ERR_RAISE(status);
 	Py_RETURN_NONE;
 }
@@ -234,19 +220,78 @@ static PyObject *py_descriptor_sacl_del(PyObject *self, PyObject *args)
 	Py_RETURN_NONE;
 }
 
+static PyObject *py_descriptor_dacl_del_ace(PyObject *self, PyObject *args)
+{
+	struct security_descriptor *desc = pytalloc_get_ptr(self);
+	NTSTATUS status;
+	struct security_ace *ace = NULL;
+	PyObject *py_ace = Py_None;
+
+	if (!PyArg_ParseTuple(args, "O!", &security_ace_Type, &py_ace))
+		return NULL;
+
+	if (!PyObject_TypeCheck(py_ace, &security_ace_Type)) {
+		PyErr_SetString(PyExc_TypeError,
+				"expected security.security_ace "
+				"for first argument to .dacl_del_ace");
+		return NULL;
+	}
+
+	ace = pytalloc_get_ptr(py_ace);
+	status = security_descriptor_dacl_del_ace(desc, ace);
+	PyErr_NTSTATUS_IS_ERR_RAISE(status);
+	Py_RETURN_NONE;
+}
+
+static PyObject *py_descriptor_sacl_del_ace(PyObject *self, PyObject *args)
+{
+	struct security_descriptor *desc = pytalloc_get_ptr(self);
+	NTSTATUS status;
+	struct security_ace *ace = NULL;
+	PyObject *py_ace = Py_None;
+
+	if (!PyArg_ParseTuple(args, "O!", &security_ace_Type, &py_ace))
+		return NULL;
+
+	if (!PyObject_TypeCheck(py_ace, &security_ace_Type)) {
+		PyErr_SetString(PyExc_TypeError,
+				"expected security.security_ace "
+				"for first argument to .sacl_del_ace");
+		return NULL;
+	}
+
+	ace = pytalloc_get_ptr(py_ace);
+	status = security_descriptor_sacl_del_ace(desc, ace);
+	PyErr_NTSTATUS_IS_ERR_RAISE(status);
+	Py_RETURN_NONE;
+}
+
 static PyObject *py_descriptor_new(PyTypeObject *self, PyObject *args, PyObject *kwargs)
 {
 	return pytalloc_steal(self, security_descriptor_initialise(NULL));
 }
 
-static PyObject *py_descriptor_from_sddl(PyObject *self, PyObject *args)
+static PyObject *py_descriptor_from_sddl(PyObject *self, PyObject *args, PyObject *kwargs)
 {
+	TALLOC_CTX *tmp_ctx = NULL;
+	static const char *kwnames[] = { "", "", "allow_device_in_sddl", NULL };
 	struct security_descriptor *secdesc;
 	char *sddl;
 	PyObject *py_sid;
+	int allow_device_in_sddl = 1;
 	struct dom_sid *sid;
+	const char *err_msg = NULL;
+	size_t err_msg_offset = 0;
+	enum ace_condition_flags ace_condition_flags = 0;
 
-	if (!PyArg_ParseTuple(args, "sO!", &sddl, &dom_sid_Type, &py_sid))
+	if (!PyArg_ParseTupleAndKeywords(args,
+					 kwargs,
+					 "sO!|$p",
+					 discard_const_p(char *, kwnames),
+					 &sddl,
+					 &dom_sid_Type,
+					 &py_sid,
+					 &allow_device_in_sddl))
 		return NULL;
 
 	if (!PyObject_TypeCheck(py_sid, &dom_sid_Type)) {
@@ -258,12 +303,54 @@ static PyObject *py_descriptor_from_sddl(PyObject *self, PyObject *args)
 
 	sid = pytalloc_get_ptr(py_sid);
 
-	secdesc = sddl_decode(NULL, sddl, sid);
-	if (secdesc == NULL) {
-		PyErr_SetString(PyExc_TypeError, "Unable to parse SDDL");
+	if (allow_device_in_sddl) {
+		ace_condition_flags |= ACE_CONDITION_FLAG_ALLOW_DEVICE;
+	}
+
+	tmp_ctx = talloc_new(NULL);
+	if (tmp_ctx == NULL) {
+		PyErr_NoMemory();
 		return NULL;
 	}
 
+	secdesc = sddl_decode_err_msg(tmp_ctx, sddl, sid,
+				      ace_condition_flags,
+				      &err_msg, &err_msg_offset);
+	if (secdesc == NULL) {
+		PyObject *exc = NULL;
+		if (err_msg == NULL) {
+			err_msg = "unknown error";
+		}
+		/*
+		 * Some notes about this exception value:
+		 *
+		 * We don't want to add the offset first, so as not to
+		 * confuse those who are used to the integer error
+		 * code coming first.
+		 *
+		 * The errant sddl is added so that the exception can
+		 * be caught some distance away from the call and we
+		 * still know what the messages refer to.
+		 */
+		exc = Py_BuildValue("(s, s, i, s)",
+				    "Unable to parse SDDL",
+				    err_msg,
+				    err_msg_offset,
+				    sddl);
+		if (exc == NULL) {
+			talloc_free(tmp_ctx);
+			/* an exception was set by Py_BuildValue() */
+			return NULL;
+		}
+		PyErr_SetObject(PyExc_SDDLValueError, exc);
+		Py_DECREF(exc);
+		talloc_free(tmp_ctx);
+		return NULL;
+	}
+
+	secdesc = talloc_steal(NULL, secdesc);
+	talloc_free(tmp_ctx);
+
 	return pytalloc_steal((PyTypeObject *)self, secdesc);
 }
 
@@ -284,6 +371,10 @@ static PyObject *py_descriptor_as_sddl(PyObject *self, PyObject *args)
 		sid = NULL;
 
 	text = sddl_encode(NULL, desc, sid);
+	if (text == NULL) {
+		PyErr_SetString(PyExc_ValueError, "Unable to encode SDDL");
+		return NULL;
+	}
 
 	ret = PyUnicode_FromString(text);
 
@@ -302,7 +393,11 @@ static PyMethodDef py_descriptor_extra_methods[] = {
 		NULL },
 	{ "sacl_del", (PyCFunction)py_descriptor_sacl_del, METH_VARARGS,
 		NULL },
-	{ "from_sddl", (PyCFunction)py_descriptor_from_sddl, METH_VARARGS|METH_CLASS, 
+	{ "dacl_del_ace", (PyCFunction)py_descriptor_dacl_del_ace, METH_VARARGS,
+		NULL },
+	{ "sacl_del_ace", (PyCFunction)py_descriptor_sacl_del_ace, METH_VARARGS,
+		NULL },
+	{ "from_sddl", (PyCFunction)py_descriptor_from_sddl, METH_VARARGS|METH_KEYWORDS|METH_CLASS,
 		NULL },
 	{ "as_sddl", (PyCFunction)py_descriptor_as_sddl, METH_VARARGS,
 		NULL },
@@ -384,7 +479,7 @@ static PyObject *py_token_is_anonymous(PyObject *self,
 	PyObject *Py_UNUSED(ignored))
 {
 	struct security_token *token = pytalloc_get_ptr(self);
-	
+
 	return PyBool_FromLong(security_token_is_anonymous(token));
 }
 
@@ -392,7 +487,7 @@ static PyObject *py_token_is_system(PyObject *self,
 	PyObject *Py_UNUSED(ignored))
 {
 	struct security_token *token = pytalloc_get_ptr(self);
-	
+
 	return PyBool_FromLong(security_token_is_system(token));
 }
 
@@ -400,7 +495,7 @@ static PyObject *py_token_has_builtin_administrators(PyObject *self,
 	PyObject *Py_UNUSED(ignored))
 {
 	struct security_token *token = pytalloc_get_ptr(self);
-	
+
 	return PyBool_FromLong(security_token_has_builtin_administrators(token));
 }
 
@@ -408,7 +503,7 @@ static PyObject *py_token_has_nt_authenticated_users(PyObject *self,
 	PyObject *Py_UNUSED(ignored))
 {
 	struct security_token *token = pytalloc_get_ptr(self);
-	
+
 	return PyBool_FromLong(security_token_has_nt_authenticated_users(token));
 }
 
@@ -437,8 +532,17 @@ static PyObject *py_token_set_privilege(PyObject *self, PyObject *args)
 
 static PyObject *py_token_new(PyTypeObject *self, PyObject *args, PyObject *kwargs)
 {
-	return pytalloc_steal(self, security_token_initialise(NULL));
-}	
+	int evaluate_claims = CLAIMS_EVALUATION_INVALID_STATE;
+	const char *kwnames[] = { "evaluate_claims", NULL };
+
+	if (!PyArg_ParseTupleAndKeywords(args, kwargs, "|i",
+					 discard_const_p(char *, kwnames),
+					 &evaluate_claims)) {
+		return NULL;
+	}
+
+	return pytalloc_steal(self, security_token_initialise(NULL, evaluate_claims));
+}
 
 static PyMethodDef py_token_extra_methods[] = {
 	{ "is_sid", (PyCFunction)py_token_is_sid, METH_VARARGS,
@@ -447,7 +551,7 @@ static PyMethodDef py_token_extra_methods[] = {
 	{ "has_sid", (PyCFunction)py_token_has_sid, METH_VARARGS,
 		NULL },
 	{ "is_anonymous", (PyCFunction)py_token_is_anonymous, METH_NOARGS,
-		"S.is_anonymus() -> bool\n"
+		"S.is_anonymous() -> bool\n"
 		"Check whether this is an anonymous token." },
 	{ "is_system", (PyCFunction)py_token_is_system, METH_NOARGS,
 		NULL },
@@ -501,10 +605,12 @@ static PyObject *py_random_sid(PyObject *self,
 {
 	struct dom_sid *sid;
 	PyObject *ret;
-    	char *str = talloc_asprintf(NULL, "S-1-5-21-%u-%u-%u", 
-			(unsigned)generate_random(), 
-			(unsigned)generate_random(), 
-			(unsigned)generate_random());
+	char *str = talloc_asprintf(
+		NULL,
+		"S-1-5-21-%"PRIu32"-%"PRIu32"-%"PRIu32,
+		generate_random(),
+		generate_random(),
+		generate_random());
 
         sid = dom_sid_parse_talloc(NULL, str);
 	talloc_free(str);
@@ -519,14 +625,119 @@ static PyMethodDef py_mod_security_extra_methods[] = {
 	{0}
 };
 
-static void py_mod_security_patch(PyObject *m)
+static bool py_mod_security_patch(PyObject *m)
 {
+	int ret;
 	int i;
 	for (i = 0; py_mod_security_extra_methods[i].ml_name; i++) {
 		PyObject *descr = PyCFunction_New(&py_mod_security_extra_methods[i], NULL);
-		PyModule_AddObject(m, py_mod_security_extra_methods[i].ml_name,
-				   descr);
+		ret = PyModule_AddObject(m, py_mod_security_extra_methods[i].ml_name,
+					 descr);
+		if (ret != 0) {
+			return false;
+		}
 	}
+	/*
+	 * I wanted to make this a subclass of ValueError, but it
+	 * seems there isn't an easy way to do that using the API.
+	 * (c.f. SimpleExtendsException in cpython:Objects/exceptions.c)
+	 */
+	PyExc_SDDLValueError = PyErr_NewException("security.SDDLValueError",
+						  NULL, NULL);
+
+	if (PyExc_SDDLValueError == NULL) {
+		return false;
+	}
+	ret = PyModule_AddObject(m, "SDDLValueError", PyExc_SDDLValueError);
+	if (ret != 0) {
+		return false;
+	}
+	return true;
 }
 
-#define PY_MOD_SECURITY_PATCH py_mod_security_patch
+#define PY_MOD_SECURITY_PATCH(m)			\
+	do {						\
+		bool _ok = py_mod_security_patch(m);	\
+		if (! _ok) {				\
+			Py_XDECREF(m);			\
+			return NULL;			\
+		}					\
+	} while(0)
+
+static PyObject *py_security_ace_equal(PyObject *py_self, PyObject *py_other, int op)
+{
+	struct security_ace *self = pytalloc_get_ptr(py_self);
+	struct security_ace *other = NULL;
+	bool eq;
+
+	if (!PyObject_TypeCheck(py_other, &security_ace_Type)) {
+		eq = false;
+	} else {
+		other = pytalloc_get_ptr(py_other);
+		eq = security_ace_equal(self, other);
+	}
+
+	switch(op) {
+	case Py_EQ:
+		if (eq) {
+			Py_RETURN_TRUE;
+		} else {
+			Py_RETURN_FALSE;
+		}
+		break;
+	case Py_NE:
+		if (eq) {
+			Py_RETURN_FALSE;
+		} else {
+			Py_RETURN_TRUE;
+		}
+		break;
+	default:
+		break;
+	}
+
+	Py_RETURN_NOTIMPLEMENTED;
+}
+
+static PyObject *py_security_ace_as_sddl(PyObject *self, PyObject *args)
+{
+	struct security_ace *ace = pytalloc_get_ptr(self);
+	PyObject *py_sid = Py_None;
+	struct dom_sid *sid = NULL;
+	char *text = NULL;
+	PyObject *ret = Py_None;
+
+	if (!PyArg_ParseTuple(args, "O!", &dom_sid_Type, &py_sid))
+		return NULL;
+
+	if (!PyObject_TypeCheck(py_sid, &dom_sid_Type)) {
+		PyErr_SetString(PyExc_TypeError,
+				"expected security.dom_sid "
+				"for second argument to .sddl_encode_ace");
+		return NULL;
+	}
+
+	sid = pytalloc_get_ptr(py_sid);
+
+	text = sddl_encode_ace(NULL, ace, sid);
+	if (text == NULL) {
+		return NULL;
+	}
+	ret = PyUnicode_FromString(text);
+	talloc_free(text);
+
+	return ret;
+}
+
+static PyMethodDef py_security_ace_extra_methods[] = {
+	{ "as_sddl", (PyCFunction)py_security_ace_as_sddl, METH_VARARGS, NULL },
+	{0}
+};
+
+#define PY_ACE_PATCH py_security_ace_patch
+
+static void py_security_ace_patch(PyTypeObject *type)
+{
+	type->tp_richcompare = py_security_ace_equal;
+	PyType_AddMethods(type, py_security_ace_extra_methods);
+}
